#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/timer.h>
#include <sys/return_code.h>
#include <sys/process.h>
#include <cell/audio.h>
#include <cell/sysmodule.h>
#include <cell/cell_fs.h>
#include <cell/pad.h>
#include <cell/dbgfont.h>
#include <stddef.h>
#include <math.h>
#include <sysutil/sysutil_sysparam.h>

#include "emulator.h"
#include "menu.h"

#include "snes9x.h"
#include "memmap.h"
#include "apu/apu.h"
#include "gfx.h"
#include "snapshot.h"
#include "controls.h"
#include "cheats.h"
#include "movie.h"
#include "logger.h"
#include "display.h"
#include "conffile.h"

#include "cellpadutil.h"
#include "cellgraphicsutil.h"
//#include "cellgraphicspsglutil.h"

#include "emulator_implementation.h"

SYS_PROCESS_PARAM(1001, 0x10000);


// mode the main loop is in
Emulator_Modes mode_switch = MODE_MENU;

// is a ROM running
bool is_running;

// is emulator loaded?
bool emulator_loaded = false;

// needs settings loaded
bool load_settings = true;

// need to load the current rom
bool need_load_rom = false;

// current rom being emulated
char* current_rom = NULL;

// controller mode to run emulator in
static Emulator_ControllerModes controller_settings = NO_JOYSTICKS;


// debug text for output during emulation loop
const char *debug_text;


void Emulator_SetControllerMode(Emulator_ControllerModes m)
{
	controller_settings = m;
}


Emulator_ControllerModes Emulator_GetControllerMode()
{
	return controller_settings;
}


bool Emulator_IsInitialized()
{
	return emulator_loaded;
}


bool Emulator_IsROMLoaded()
{
	return current_rom != NULL && need_load_rom == false;
}


bool Emulator_ROMRunning()
{
	return is_running;
}

char * Emulator_GetROM()
{
	return current_rom;
}

void Emulator_SetDebugMessage(const char * value)
{
	debug_text = value;
}


void Emulator_SwitchMode(Emulator_Modes m)
{
	mode_switch = m;
}


void Emulator_Shutdown()
{
	cellDbgFontExitGcm();
	cellSysmoduleUnloadModule(CELL_SYSMODULE_FS);
	cellSysmoduleUnloadModule(CELL_SYSMODULE_IO);
	cellSysmoduleUnloadModule(CELL_SYSMODULE_AUDIO);
	cellSysutilUnregisterCallback(0);


	// do any clean up... save stuff etc
	// ...

	// force exit
	exit(0);
}


void Emulator_RequestLoadROM(char* rom, bool forceReload)
{
	if (forceReload || current_rom == NULL || strcmp(rom, current_rom) != 0)
	{
		if (current_rom != NULL)
		{
			free(current_rom);
		}

		current_rom = strdup(rom);
		need_load_rom = true;
	}
	else
	{
		need_load_rom = false;
	}
}

void Emulator_SaveSRAM()
{
	S9xSaveSRAM();
}


void Emulator_StopROMRunning()
{
	is_running = false;
}

void Emulator_StartROMRunning()
{
	Emulator_SwitchMode(MODE_EMULATION);
}

/*
 * Callback for PS3 System operations
 */
void sysutil_exit_callback (uint64_t status, uint64_t param, void *userdata) {
	(void) param;
	(void) userdata;

	switch (status) {
		case CELL_SYSUTIL_REQUEST_EXITGAME:
			MenuStop();
			Emulator_StopROMRunning();
			Emulator_SwitchMode(MODE_EXIT);
			break;
		case CELL_SYSUTIL_DRAWING_BEGIN:
		case CELL_SYSUTIL_DRAWING_END:
			break;
	}
}


bool Emulator_InitSettings ()
{
	//new function added to emulator_implementation; also had to create emulator_implementation.h to add the function
	S9xInitSettings();
}


bool Emulator_Init()
{
	return S9xInitSystem(controller_settings);
}

bool Emulator_Initialize()
{
	emulator_loaded = Emulator_Init();
}


void Emulator_Inputloop () {
	uint8_t pads_connected = cellPadUtilNumberPadsConnected();
	for (uint8_t i = 0; i < pads_connected; ++i) {
		if (cellPadUtilUpdate(i))
		{
			S9xReportPad(i);
		}
	}
}

static uint32_t audio_port;
static float *audio_buffer;
static float *audio_buffer_end;
sys_addr_t audio_read_index_addr;
static unsigned int audio_block_index;
static unsigned int audio_last_read_block;

void init_sound () {
	CellAudioPortParam params;
	CellAudioPortConfig config;

	cellAudioInit();

	params.nChannel = AUDIO_CHANNEL_COUNT;
	params.nBlock = AUDIO_BLOCK_COUNT;
	params.attr = 0;

	if (cellAudioPortOpen(&params, &audio_port) != CELL_OK) {
		cellAudioQuit();
		return;
	}

	cellAudioGetPortConfig(audio_port, &config);
	audio_buffer = (float *)config.portAddr;
	audio_buffer_end = (float *)(config.portAddr + config.portSize);
	audio_read_index_addr = config.readIndexAddr;

	cellAudioPortStart(audio_port);
}

void make_noise () {
	int16_t samples[256 * 2 * 2];

	if (S9xGetSampleCount() < 512)
		return;

	unsigned int current_block = *(uint64_t *)audio_read_index_addr;

	if (current_block == audio_last_read_block)
		return;

	while (S9xGetSampleCount() >= 512 && audio_block_index != current_block) {
		S9xMixSamples((uint8 *)samples, 256 * 2);

		float *buf = audio_buffer + AUDIO_CHANNEL_COUNT * 256 * audio_block_index;
		for (unsigned int i = 0; i < 256 * AUDIO_CHANNEL_COUNT; ++i)
			buf[i] = ((float)samples[i])/32768.0f;

		audio_block_index = (audio_block_index + 1) % AUDIO_BLOCK_COUNT;
	}

	audio_last_read_block = current_block;
}



void Emulator_Start()
{
	//S9X_DBG("if(need_load_rom)");
	if(need_load_rom)
	{
		//S9X_DBG("if(need_load_rom) == true");
		if (!Emulator_Initialize())	
			Settings.Paused = TRUE;
		else
			need_load_rom = false;
	}

	
	memset(display, 0, color_buf[0].height * color_buf[0].pitch);
	memset(display_back, 0, color_buf[0].height * color_buf[0].pitch);
	//S9X_DBG("S9xSetSoundMute(FALSE)");
	S9xSetSoundMute(FALSE);


	//S9X_DBG("while (is_running)");
	while (is_running) {
		//S9X_DBG("while (is_running) == true");
		//S9X_DBG("Emulator_Inputloop()");
		Emulator_Inputloop();

		if (!Settings.Paused) {
			//S9X_DBG("Make noise");
			make_noise();
			//S9X_DBG("if (S9xSyncSound)");
			if (S9xSyncSound())
				//S9X_DBG("if (S9xSyncSound) == true)");
				//S9X_DBG("S9xMainLoop");
				S9xMainLoop();
				//S9X_DBG("S9xMainLoop done");
		}

		/*
		if (debug_text)
			debugOutStr(100, 100, debug_text);
			*/

		cellGcmUtilRender(false);
	}


	// will always default to true
	// if (Emulator_IsInitialized())
	Emulator_SaveSRAM();
	S9xSetSoundMute(TRUE);
	is_running = true;

}

int main (void) {
	cellSysmoduleLoadModule(CELL_SYSMODULE_FS);
	cellSysmoduleLoadModule(CELL_SYSMODULE_IO);
	cellSysmoduleLoadModule(CELL_SYSMODULE_AUDIO);

	cellSysutilRegisterCallback(0, sysutil_exit_callback, NULL);

	cellPadUtilPadInit();
	if (cellGcmUtilInitDisplay() != CELL_OK)
	{
		Emulator_Shutdown();
	}
	//cellGcmUtilInitShader();
	//cellPSGLUtilInitDisplay();
	//S9X_DBG("cellPSGLUtilInitDisplay");
	is_running = true;
	//cellGcmUtilInitDisplay2();
	//S9X_DBG("cellGcmUtilInitDbgFont()")
	/*if(cellGcmUtilInitDbgFont() == FALSE)
		exit(0);*/
	//S9X_DBG("init_sound");
	init_sound();

	// load SNES9x settings before calling rom_menu() so they are displayed
	// correctly, in case settings screen is accessed before a ROM is loaded

	if (load_settings)
	{
		//S9X_DBG("Emulator_InitSettings");
		Emulator_InitSettings();
		load_settings =  false;
	}

	//S9X_DBG("Emulator_SwitchMode(MODE_EMULATION)");
	//Emulator_SwitchMode(MODE_EMULATION);
	//S9X_DBG("Emulator_SwitchMode(MODE_EMULATION)");
	//S9X_DBG("Emulator_RequestLoadROM(\"/dev_hdd0/game/SNES90000/USRDIR/game.sfc\", false);");
	//Emulator_RequestLoadROM("/dev_hdd0/game/SNES90000/USRDIR/game.sfc", false);

	while(1)
	{
		//S9X_DBG("switch(mode_switch");
		switch(mode_switch)
		{
			case MODE_MENU:
				MainMenu();
				break;
			case MODE_EMULATION:
				//S9X_DBG("Emulator_Start()");
				Emulator_Start();
				break;
			case MODE_EXIT:
				Emulator_Shutdown();
				return 0;
		}
	}

	return 0;
}

